// **************************************************************************************
// el Comandante                                                    guerilla.bplaced.net
// --------------------------------------------------------------------------------------
// Information:
// Simple analog joystick test, should work with any standard analog joysticks
// For the sprite (screen back buffer) enough RAM is needed
// The TFT resolution should be 480 x 320 or greater, 240 x 135 maybe also work
// --------------------------------------------------------------------------------------
// Hardware:
// WT32-SC01 (Arduino --> ESP32 Wrover Module) / LILYGO-TTGO-T-Display --> (Arduino --> ESP32 Dev Module)
// Analog Joystick, PSP2000 / Nintendo Switch / Breakout Modul...
// Notice:
// I bought two replicas of a PSP2000 joystick at amazon, the quality is very, very bad,
// the parts are not usable! At the end a standard analog joystick breakout modul was used.
// That cheap part is far away to be perfect, but it is still usable for simple applications.
// --------------------------------------------------------------------------------------
// Libraries:
// TFT_eSPI --> https://github.com/Bodmer/TFT_eSPI
// --------------------------------------------------------------------------------------
// Documentations:
// Overview ESP32 mcu's   - https://socialcompare.com/en/comparison/esp8266-vs-esp32-vs-esp32-s2
// Espressif ESP32 docu's - https://www.espressif.com/en/support/documents/technical-documents
// --------------------------------------------------------------------------------------
// Device WT32-SC01       - The esp32 device WT32-SC01 can not read pixel from the TFT using HSPI
// Device WT32-SC01       - http://myosuploads3.banggood.com/products/20220111/20220111020446WT32-SC01DataSheetV3.3.pdf
// TFT Device ST7796      - https://cdn.hackaday.io/files/1625926956336128/ST7796S.pdf
// Touch Device FT6336U   - https://focuslcds.com/content/FT6236.pdf
// --------------------------------------------------------------------------------------
// Device TTGO-T-Display  - The esp32 device LILYGO-TTGO-T-Display can read pixel from the TFT
// Device TTGO-T-Display  - https://github.com/Xinyuan-LilyGO/TTGO-T-Display
// TFT Device ST7789V     - https://newhavendisplay.com/content/datasheets/ST7789V.pdf
// **************************************************************************************
#include <TFT_eSPI.h>                           // TFT_espi library
TFT_eSPI tft          = TFT_eSPI();
TFT_eSprite graph     = TFT_eSprite(&tft);      // sprite tft screen back buffer

// ######################################################################################
// Change the values to match your system and desires
// ######################################################################################
#define analogX_PIN  26                         // FREE esp32 pin GPIO26 (ADC2_9), x = left/right
#define analogY_PIN  25                         // FREE esp32 pin GPIO25 (ADC2_8), y = up/down
#define digitalB_PIN 36                         // FREE esp32 pin GPIO36 (SensVP), button (low active)
#define xMin 0                                  // x-axis, min. raw value from mcu analog input pin
#define xMax 4095                               // x-axis, max. raw value from mcu analog input pin (12 bit = 4095)
#define yMin 0                                  // y-axis, min. raw value from mcu analog input pin
#define yMax 4095                               // y-axis, max. raw value from mcu analog input pin (12 bit = 4095)
#define speed 0                                 // time delay in ms for joystick loop, 0 = ~ no delay

// Vars
uint8_t     textSize;                           // text size
uint32_t    AllTextX;                           // x position of all values description text
uint32_t    xAxisTextY;                         // y position of each values text
uint32_t    yAxisTextY;
uint32_t    xOffsetTextY;
uint32_t    yOffsetTextY;
uint32_t    xCenterY;
uint32_t    yCenterY;
uint32_t    xScreenTextY;
uint32_t    yScreenTextY;
uint32_t    RadiusTextY;
uint32_t    DistanceTextY;
int8_t      mode = 0;                           // mode, 0 = square, 1 = circle
int8_t      flagInit = 0;                       // flag if graphic mode (square/circle) is already init
int32_t     x;                                  // x-axis, raw value from mcu analog input pin
int32_t     y;                                  // y-axis, raw value from mcu analog input pin
int32_t     xOffset = 0;                        // x, used to calibrate joystick and zero position (not implemented)
int32_t     yOffset = 0;                        // y, used to calibrate joystick and zero position (not implemented)
int32_t     xCenter;                            // x center of square / circle
int32_t     yCenter;                            // y center of square / circle
uint32_t    xScreen;                            // x tft coordinate of joystick
uint32_t    yScreen;                            // y tft coordinate of joystick
uint32_t    radius;                             // radius of circle / half width/height of square
uint32_t    distance;                           // current distance from suare/circle center to joystick position x/y coordinates
uint32_t    xScreenOld;                         // restore old screen area
uint32_t    yScreenOld;                         // restore old screen area
uint32_t    distanceOld;                        // old distance, to check for new drawing at mode 1 (circle)
uint32_t    tftWidth;                           // tft width
uint32_t    tftHeight;                          // tft height
uint32_t    update;                             // update time

void setup()
{
  //Serial.begin(115200);                         // init serial monitor
  tft.init();                                   // TFT_espi - init library
  digitalWrite(TFT_BL, 0);                      // switch backlight off
  pinMode(digitalB_PIN, INPUT);                 // set button pin mode
  systemInit();                                 // init system
  screenInit();                                 // init screen
  graphInit();                                  // init graphic buffer
  systemInfo();                                 // display system info
  delay(200);                                   // for a coffee
  digitalWrite(TFT_BL, 1);                      // switch backlight on
}

void loop()
{
  if (update <= millis())
  {
    update = millis() + speed;                  // update time in milliseconds
    // push button to toggle mode
    if (digitalRead(digitalB_PIN) == LOW)
    {mode = mode ^ 1; flagInit = 0;}
  
    // check selected mode, check if already init, run mode
    switch (mode)
    {
      case 0:
      if (flagInit == 0)
      {graphMode0SquareInit(); flagInit = 1;}
      joyAnalogMode0SquareRun();
      break;

      case 1:
      if (flagInit == 0)
      {graphMode1CircleInit(); flagInit = 1;}
      joyAnalogMode1CircleRun();
      break;
    }
  }
}



// **************************************************************************************
// System - init
void systemInit()
{
  tft.setRotation(3);                           // set desired rotation
  tftWidth = tft.width();                       // get tft width
  tftHeight = tft.height();                     // get tft height

  if (tftWidth >= 320 || tftHeight >= 320)      // big screen (i.e. WT32-SC01 - 480 x 320)
  {textSize = 2;}
  else if (tftWidth < 320 || tftHeight < 320)   // small screen (i.e. LILYGO TTGO-T-Display - 240 x 135)
  {textSize = 1;}

  xCenter = tftWidth / 100.0 * 73;              // x circle center
  yCenter = tftHeight / 2.0;                    // y circle center
  radius  = tftHeight / 100.0 * 32;             // radius circle
  AllTextX        = tftWidth / 100.0 * 31;      // x position all text
  xAxisTextY      = tftHeight / 100.0 * 14;     // y position text
  yAxisTextY      = tftHeight / 100.0 * 21.5;
  xOffsetTextY    = tftHeight / 100.0 * 29;
  yOffsetTextY    = tftHeight / 100.0 * 36.5;
  xCenterY        = tftHeight / 100.0 * 44;
  yCenterY        = tftHeight / 100.0 * 51.5;
  xScreenTextY    = tftHeight / 100.0 * 59;
  yScreenTextY    = tftHeight / 100.0 * 66.5;
  RadiusTextY     = tftHeight / 100.0 * 74;
  DistanceTextY   = tftHeight / 100.0 * 81.5;
}

// **************************************************************************************
// Screen - init
void screenInit()
{
  tft.fillScreen(0x0000);                       // set desired background color
  tft.setTextFont(1);                           // select font
  tft.setTextSize(textSize);
  tft.setTextDatum(MC_DATUM);
  tft.setTextColor(0xFFFE, 0x0000);
}
// **************************************************************************************
// Graph - init
void graphInit()
{
  graph.setColorDepth(8);                       // 8 bit colors, to minimize required RAM
  graph.createSprite(tftWidth, tftHeight);      // set sprite dimension to tft dimension
  graph.fillSprite(0x0000);
  graph.setTextFont(1);                         // select font
  graph.setTextSize(textSize);
  graph.setTextDatum(MC_DATUM);
  graph.setTextColor(0xFFFE, 0x0000);
}

// **************************************************************************************
// Graphic - create graphic screen mode 0 (square)
void graphMode0SquareInit()
{
  digitalWrite(TFT_BL, 0);                      // switch backlight off

  graph.fillSprite(0x0000);
  graph.drawString("Selected mode 0 (Square)", tftWidth / 2.0, tftHeight / 100.0 * 5);
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 14);
  graph.println("x-Axis: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 21.5);
  graph.println("y-Axis: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 29);
  graph.println("x-Offset: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 36.5);
  graph.println("y-Offset: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 44);
  graph.println("x-Center: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 51.5);
  graph.println("y-Center: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 59);
  graph.println("x-Screen: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 66.5);
  graph.println("y-Screen: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 74);
  graph.println("Radius: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 81.5);
  graph.println("Distance:");

  graph.drawString("Push Button to toggle mode...", tftWidth / 100.0 * 50, tftHeight / 100.0 * 95);
  graph.drawLine(0, tftHeight / 100.0 * 10, tftWidth, tftHeight / 100.0 * 10, 0x8410);
  graph.drawRect(xCenter - radius, yCenter - radius, radius * 2, radius * 2, 0x9492), 
  graph.drawLine(0, tftHeight / 100.0 * 90, tftWidth, tftHeight / 100.0 * 90, 0x8410);
  graph.drawLine(xCenter - 10, tftHeight / 2, xCenter + 10, tftHeight / 2, 0xFFFE);
  graph.drawLine(xCenter, tftHeight / 2 - 10, xCenter, tftHeight / 2 + 10, 0xFFFE);

  graph.pushSprite(0, 0, 0, 0, tftWidth, tftHeight);
  delay(200);

  digitalWrite(TFT_BL, 1);                      // switch backlight on
}

// **************************************************************************************
// Graphic - create graphic screen mode 1 (circle)
void graphMode1CircleInit()
{
  digitalWrite(TFT_BL, 0);                      // switch backlight off

  graph.fillSprite(0x0000);
  graph.drawString("Selected mode 1 (Circle)", tftWidth / 2.0, tftHeight / 100.0 * 5);
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 14);
  graph.println("x-Axis: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 21.5);
  graph.println("y-Axis: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 29);
  graph.println("x-Offset: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 36.5);
  graph.println("y-Offset: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 44);
  graph.println("x-Center: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 51.5);
  graph.println("y-Center: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 59);
  graph.println("x-Screen: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 66.5);
  graph.println("y-Screen: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 74);
  graph.println("Radius: ");
  graph.setCursor(tftWidth / 100.0 * 2, tftHeight / 100.0 * 81.5);
  graph.println("Distance:");

  graph.drawString("Push Button to toggle mode...", tftWidth / 100.0 * 50, tftHeight / 100.0 * 95);
  graph.drawLine(0, tftHeight / 100.0 * 10, tftWidth, tftHeight / 100.0 * 10, 0x8410);
  graph.drawCircle(xCenter, tftHeight / 2, radius, 0x9492);
  graph.drawLine(0, tftHeight / 100.0 * 90, tftWidth, tftHeight / 100.0 * 90, 0x8410);
  graph.drawLine(xCenter - 10, tftHeight / 2, xCenter + 10, tftHeight / 2, 0xFFFE);
  graph.drawLine(xCenter, tftHeight / 2 - 10, xCenter, tftHeight / 2 + 10, 0xFFFE);

  graph.pushSprite(0, 0, 0, 0, tftWidth, tftHeight);
  delay(200);

  digitalWrite(TFT_BL, 1);                      // switch backlight on
}

// **************************************************************************************
// Joystick - run in graphic screen mode 0 (square)
void joyAnalogMode0SquareRun()
{
  // read x and y raw value from mcu analog in pin
  x = analogRead(analogX_PIN);
  y = analogRead(analogY_PIN);

  // map x and v value from mcu analog in pin to (square) circle area
  xScreen = map(x, xMin, xMax, xCenter + radius, xCenter - radius);
  yScreen = map(y, yMin, yMax, yCenter - radius, yCenter + radius);

  // distance from center x/y point of the square
  distance = sqrt(sq(xScreen - xCenter) + sq(yScreen - yCenter));

  // check if update necessary
  if (xScreen != xScreenOld || yScreen != yScreenOld)
  {
    // restore old tft area and draw at point x/y
    graph.pushSprite(xScreenOld - 10, yScreenOld - 10, xScreenOld - 10, yScreenOld - 10, 20, 20);
    tft.fillCircle(xScreen, yScreen, 6, 0xF800);
    xScreenOld = xScreen;
    yScreenOld = yScreen;
  }

  // display values
  displayValues();
}

// **************************************************************************************
// Joystick - run in graphic screen mode 1 (circle)
void joyAnalogMode1CircleRun()
{
  // read x and y raw value from mcu analog in pin
  x = analogRead(analogX_PIN);
  y = analogRead(analogY_PIN);

  // map x and v value from mcu analog in pin to (square) circle area
  xScreen = map(x, xMin, xMax, xCenter + radius, xCenter - radius);
  yScreen = map(y, yMin, yMax, yCenter - radius, yCenter + radius);

  // check if x/y jostick coordinates are inside the circle, limit movement inside circle area
  distance = sqrt(sq(xScreen - xCenter) + sq(yScreen - yCenter));
  if (distance < radius && distance != distanceOld)
  {
    distanceOld = distance;
    // restore old tft area and draw at point x/y
    graph.pushSprite(xScreenOld - 10, yScreenOld - 10, xScreenOld - 10, yScreenOld - 10, 20, 20);
    tft.fillCircle(xScreen, yScreen, 6, 0xF800);
    xScreenOld = xScreen;
    yScreenOld = yScreen;
  }

  // enable movement on circular path
  else if (distance != distanceOld && (xScreen >= xCenter - radius && xScreen <= xCenter + radius) && (yScreen >= yCenter - radius && yScreen <= yCenter + radius))
  {
    distanceOld = distance;
    if (yScreen < yCenter)
    {yScreen = yCenter - sqrt(sq(radius) - sq(xScreen - xCenter));}
    else
    {yScreen = yCenter + sqrt(sq(radius) - sq(xScreen - xCenter));}

    // restore old tft area and draw at point x/y
    graph.pushSprite(xScreenOld - 10, yScreenOld - 10, xScreenOld - 10, yScreenOld - 10, 20, 20);
    tft.fillCircle(xScreen, yScreen, 6, 0xF800);
    xScreenOld = xScreen;
    yScreenOld = yScreen;
  }

  // display values
  displayValues();
}

// **************************************************************************************
// Display values
void displayValues()
{
  // display values
  tft.setCursor(AllTextX, xAxisTextY);
  tft.printf("%d   ", x);
  tft.setCursor(AllTextX, yAxisTextY);
  tft.printf("%d   ", y);
  tft.setCursor(AllTextX, xOffsetTextY);
  tft.printf("N/A   ");
  tft.setCursor(AllTextX, yOffsetTextY);
  tft.printf("N/A   ");
  tft.setCursor(AllTextX, xCenterY);
  tft.printf("%d   ", xCenter);
  tft.setCursor(AllTextX, yCenterY);
  tft.printf("%d   ", yCenter);
  tft.setCursor(AllTextX, xScreenTextY);
  tft.printf("%d   ", xScreen);
  tft.setCursor(AllTextX, yScreenTextY);
  tft.printf("%d   ", yScreen);
  tft.setCursor(AllTextX, RadiusTextY);
  tft.printf("%d   ", radius);
  tft.setCursor(AllTextX, DistanceTextY);
  tft.printf("%d   ", distance);
}

// **************************************************************************************
// System - info
void systemInfo()
{
  digitalWrite(TFT_BL, 0);                      // switch backlight off

  esp_chip_info_t chip_info;
  esp_chip_info(&chip_info);

  tft.drawString("System Information", tftWidth / 2, tftHeight / 100.0 * 5);
  tft.drawLine(0, tftHeight / 100.0 * 10, tftWidth, tftHeight / 100.0 * 10, 0x8410);

  tft.setCursor(16, tftHeight / 100.0 * 16);
  tft.printf("%s", (chip_info.model & CHIP_ESP32) ?   "" : "Model:          ESP32");
  tft.printf("%s", (chip_info.model & CHIP_ESP32S2) ? "" : "Model:          ESP32-S2");
  tft.printf("%s", (chip_info.model & CHIP_ESP32S3) ? "" : "Model:          ESP32-S3");
  tft.printf("%s", (chip_info.model & CHIP_ESP32C3) ? "" : "Model:          ESP32-C3");
  tft.setCursor(16, tftHeight / 100.0 * 24);
  tft.printf("Cores:          %d", chip_info.cores);
  tft.setCursor(16, tftHeight / 100.0 * 32);
  tft.printf("Revision:       %d", chip_info.revision);
  tft.setCursor(16, tftHeight / 100.0 * 40);
  tft.printf("Flash RAM:      %d", ESP.getFlashChipSize());   // print FLASH RAM size
  tft.setCursor(16, tftHeight / 100.0 * 48);
  tft.printf("SP RAM:         %d", esp_spiram_get_size());    // print PS RAM size
  tft.setCursor(16, tftHeight / 100.0 * 56);
  tft.printf("Free Heap Size: %d", esp_get_free_heap_size()); // print Heap size
  tft.setCursor(16, tftHeight / 100.0 * 64);
  tft.printf("WLAN:           %s", (chip_info.features & CHIP_FEATURE_WIFI_BGN) ? "2.4 GHz" : "-");
  tft.setCursor(16, tftHeight / 100.0 * 73);
  tft.printf("Bluetooth:      %s%s", (chip_info.features & CHIP_FEATURE_BT) ? "BT" : "-", (chip_info.features & CHIP_FEATURE_BLE) ? "/BLE" : "");
  tft.setCursor(16, tftHeight / 100.0 * 80);
  tft.printf("IEEE 802.15.4:  %s", (chip_info.features & CHIP_FEATURE_IEEE802154) ? "IEEE 802.15.4" : "-");

  tft.drawLine(0, tftHeight / 100.0 * 90, tftWidth, tftHeight / 100.0 * 90, 0x8410);
  tft.drawString("Push Button to continue...", tftWidth / 2, tftHeight / 100.0 * 95);

  digitalWrite(TFT_BL, 1);                      // switch backlight on

  while (digitalRead(digitalB_PIN) != LOW)
  {}
}
